from datetime import timedelta, datetime
from abc import ABC, abstractmethod
from typing import Dict, Any
import numpy

from dataclasses import dataclass
from .test_functions import (sphere, rastrigin, ackley,
                             rosenbrock, griewank, schwefel,
                             levy, michalewicz, zakharov,
                             dixonprice, st)


@dataclass
class Result:
    test_function: str
    success_percent: float
    calls: float
    input_args: Dict[str, Any]
    iterations: int
    avg_time: str
    avg_iterations: float


class BaseBenchMark(ABC):

    method_class: type = None
    start_method_name: str = None
    method_kwargs: dict = None
    test_function: callable = None
    test_function_name: str = None
    function_attr: str = None

    def __init__(self, method_class: type, start_method_name: str,
                 method_kwargs: dict, x_min: list, x_max: list,
                 n: int = 2, iterations: int = 100,
                 x_eps=0.001, function_attr: str = 'f', x_min_argument='x_min',
                 x_max_argument='x_max'):
        self.method_class = method_class
        self.start_method_name = start_method_name
        self.method_kwargs = method_kwargs
        self.function_attr = function_attr
        self.x_min = x_min
        self.x_max = x_max
        self.x_min_argument = x_min_argument
        self.x_max_argument = x_max_argument
        self.method = self.method_class(
            **self.method_kwargs,
            **{
                self.function_attr: self.test_function,
                self.x_min_argument: self.x_min,
                self.x_max_argument: self.x_max
            },
        )
        self.n = n
        self.iterations = iterations
        self.x_eps = x_eps

    def __call__(self) -> Result:
        for attr in {'calls', 'reset'}:
            if not hasattr(self.test_function, attr):
                raise AttributeError(f"Отсутствует обязательный атрибут у тестовой функции: {attr}")
        successes = 0
        calls_count = 0
        iterations_count = 0
        sum_timedelta = timedelta(0)
        for i in range(self.iterations):
            t = datetime.now()
            res_x, res_f, res_it = getattr(self.method, self.start_method_name)(self.canonical_x, self.x_eps)
            iterations_count += res_it
            sum_timedelta += datetime.now() - t
            if numpy.linalg.norm(res_x - self.canonical_x) < self.x_eps:
                successes += 1
            calls_count += self.test_function.calls
            self.test_function.reset()
            self.reset_method()
        result = Result(
            success_percent=100 * successes / self.iterations,
            calls=calls_count / self.iterations,
            input_args={
                **self.method_kwargs,
                **{
                    self.x_min_argument: self.x_min,
                    self.x_max_argument: self.x_max
                }
            },
            test_function=self.test_function_name,
            iterations=self.iterations,
            avg_time=str(sum_timedelta / self.iterations),
            avg_iterations=iterations_count / self.iterations
        )
        return result

    def reset_method(self):
        self.method = self.method_class(
            **self.method_kwargs,
            **{
                self.function_attr: self.test_function,
                self.x_min_argument: self.x_min,
                self.x_max_argument: self.x_max
            },
        )

    @property
    @abstractmethod
    def canonical_x(self) -> numpy.array:
        pass


class SphereBenchMark(BaseBenchMark):

    test_function = sphere
    test_function_name = 'Sphere'

    @property
    def canonical_x(self) -> numpy.array:
        return numpy.array(self.n * [0])


class RastriginBenchMark(BaseBenchMark):

    test_function = rastrigin
    test_function_name = 'Rastrigin'

    @property
    def canonical_x(self) -> numpy.array:
        return numpy.array(self.n * [0])


class AckleyBenchMark(BaseBenchMark):

    test_function = ackley
    test_function_name = 'Ackley'

    @property
    def canonical_x(self) -> numpy.array:
        return numpy.array(self.n * [0])


class RosenbrockBenchMark(BaseBenchMark):

    test_function = rosenbrock
    test_function_name = 'Rosenbrock'

    @property
    def canonical_x(self) -> numpy.array:
        return numpy.array(self.n * [1])


class GriewankBenchMark(BaseBenchMark):

    test_function = griewank
    test_function_name = 'Griewank'

    @property
    def canonical_x(self) -> numpy.array:
        return numpy.array(self.n * [0])


class SchwefelBenchMark(BaseBenchMark):

    test_function = schwefel
    test_function_name = 'Schwefel'

    @property
    def canonical_x(self) -> numpy.array:
        return numpy.array(self.n * [420.9687])


class LevyBenchMark(BaseBenchMark):

    test_function = levy
    test_function_name = 'Levy'

    @property
    def canonical_x(self) -> numpy.array:
        return numpy.array(self.n * [1])


# class MichalewiczBenchMark(BaseBenchMark):
#
#     test_function = michalewicz
#     test_function_name = 'Michalewicz'
#
#     @property
#     def canonical_x(self) -> numpy.array:
#         return numpy.array(self.n * [0])


class ZakharovBenchMark(BaseBenchMark):

    test_function = zakharov
    test_function_name = 'Zakharov'

    @property
    def canonical_x(self) -> numpy.array:
        return numpy.array(self.n * [0])


class DixonpriceBenchMark(BaseBenchMark):

    test_function = dixonprice
    test_function_name = 'Dixonprice'

    @property
    def canonical_x(self) -> numpy.array:
        return numpy.array([2 ** (-(2 ** i - 2) / 2 ** i) for i in range(1, self.n + 1)])


class StBenchMark(BaseBenchMark):
    test_function = st
    test_function_name = 'Styblinski–Tang'

    @property
    def canonical_x(self) -> numpy.array:
        return numpy.array(self.n * [-2.903534])



from .country_optimization import CountriesAlgorithm
from .country_optimization_v2 import CountriesAlgorithm_v2
from .country_optimization_v3 import CountriesAlgorithm as CountriesAlgorithm_v3
from .genetic_optimization import GeneticAlgorithm
from .iwo import IWO
from .gray_wolves import GreyWolfOptimizer
from .bees import BeesOptimizer
from .sand_cats import SandCatOptimizer



from krashemit.algorithms import GeneticAlgorithm
from krashemit.algorithms import CountriesAlgorithm
from krashemit.algorithms import CountriesAlgorithm_v2
from krashemit.algorithms import CountriesAlgorithm_v3
from krashemit.benchmakrs.benchmark import SphereBenchMark
from krashemit.benchmakrs.aggregators import BenchMarkParamsAggregator
from krashemit.benchmakrs.plotters import BenchMarkParamsPlotter
from krashemit.benchmakrs.tg import TgSender


# benchmark = StBenchMark(
#     method_class=GeneticAlgorithm,
#     start_method_name='testing',
#     method_kwargs={
#         'n': 200,
#         'child_percent': 20,
#         'mutation_chance': 25,
#         'max_mutation': 8,
#         # 'x_min': 10 * [-5.12],
#         # 'x_max': 10 * [5.12],
#         'genes': 5 * [16],
#         't_max': 100
#     },
#     n=5,
#     iterations=10,
#     x_min=[-5.12, -5.12, -5.12, -5.12, -5.12],
#     x_max=[5.12, 5.12, 5.12, 5.12, 5.12]
# )
# res = benchmark()
# 	M=10,
# 	N=25,
# 	n=[1, 10],
# 	p=[0.00001, 2],
# 	m=[1, 8],
# 	k=8,
# 	l=3,
# 	ep=[0.2, 0.4],
aggr = BenchMarkParamsAggregator(
    method_class=CountriesAlgorithm_v2,
    start_method_name='testing',
    method_kwargs={
        'M': 5,
        'n': [2, 5],
        'p': [0.00001, 2],
        'm': [1, 3],
        'k': 3,
        'l': 1,
        'ep': [0.2, 0.4],
        'tmax': 100
    },
    param_name='N',
    params_values=[10, 15, 20, 25],
    benchmarks={
        'sphere':
            {
                'class': SphereBenchMark,
                'params': {
                    'n': 5,
                    'iterations': 50,
                    'x_min': 5 * [-5],
                    'x_max': 5 * [5],
                    'x_min_argument': 'Xmin',
                    'x_max_argument': 'Xmax',
                }
            },
    }
)
res = aggr()
plotter = BenchMarkParamsPlotter(res, 'N', 'АВС1', path='C:/Users/Roman/PycharmProjects/KrasheMit/docs/')
plot_names = plotter.plot()
print(plot_names)
sender = TgSender(
    bot_token='5642427180:AAGU0AR0ONiaym-8edcFmZja9dwdOzGZ458',
    chat_id='247194445',
    results=res,
    x_param='N',
    method_name='АВС1',
    plot_names=plot_names,
    path='C:/Users/Roman/PycharmProjects/KrasheMit/docs/'
)
sender.send()
print(res)



import copy
import numpy as np
from typing import Callable, Iterable


class Wolf:

    def __init__(self, x, function):
        self.x = np.array(x)
        self.f = function(self.x)
        self.function = function

    def update_wolf(self, a: float, alpha, beta, delta, ub, lb):
        r1 = np.random.random(self.x.size)
        r2 = np.random.random(self.x.size)

        A1 = 2 * a * r1 - a
        C1 = 2 * r2
        D_alpha = np.abs(C1 * alpha.x - self.x)
        X1 = alpha.x - A1 * D_alpha

        r1 = np.random.random(self.x.size)
        r2 = np.random.random(self.x.size)
        A2 = 2 * a * r1 - a
        C2 = 2 * r2
        D_beta = np.abs(C2 * beta.x - self.x)
        X2 = beta.x - A2 * D_beta

        r1 = np.random.random(self.x.size)
        r2 = np.random.random(self.x.size)
        A3 = 2 * a * r1 - a
        C3 = 2 * r2
        D_delta = np.abs(C3 * delta.x - self.x)
        X3 = delta.x - A3 * D_delta
        self.x = (X1 + X2 + X3) / 3
        self.x = np.clip(self.x, lb, ub)
        self.f = self.function(self.x)

    def __lt__(self, other):
        return self.f < other.f

    def __le__(self, other):
        return self.f <= other.f

    def __gt__(self, other):
        return self.f > other.f

    def __ge__(self, other):
        return self.f >= other.f

class GreyWolfOptimizer:

    def __init__(
            self,
            f: Callable,
            x_min: Iterable[float],
            x_max: Iterable[float],
            n_wolves: int = 30,
            max_iter: int = 500,
            printing=False
    ):
        self.objective_func = f
        self.n_wolves = n_wolves
        self.max_iter = max_iter

        self.lb = np.array(x_min)
        self.ub = np.array(x_max)

        self.printing = printing

    def _first_iteration(self):
        self.t = 0
        self.wolves = []
        for i in range(self.n_wolves):
            xi = np.random.uniform(self.ub, self.lb)
            self.wolves.append(Wolf(xi, self.objective_func))
        self.wolves.sort()

    def _iteration(self):
        alpha, beta, delta = tuple(map(copy.copy, self.wolves[:3]))
        a = 2 - self.t * (2 / self.max_iter)
        for wolf in self.wolves:
            wolf.update_wolf(a, alpha, beta, delta, self.ub, self.lb)
        self.wolves.sort()
        if self.printing:
            print(self.t, self.wolves[0].x, self.wolves[0].f)
        self.t += 1

    def start(self):
        self._first_iteration()
        while self.t <= self.max_iter:
            stop = self._iteration()
            if stop:
                break
        return (self.wolves[0].x, self.wolves[0].f, False, self.t)

    def testing(self, canonical_x, epsilon):
        self._first_iteration()
        while self.t <= self.max_iter:
            self._iteration()
            if np.linalg.norm(canonical_x - np.array(self.wolves[0].x)) < epsilon:
                return self.wolves[0].x, self.wolves[0].f, self.t
        return self.wolves[0].x, self.wolves[0].f, self.t



import random
import numpy as np
from math import ceil


def random_matrix(x, x_min, x_max):
    return x * np.random.uniform(x_min, x_max)

def check_population(country):
    return bool(country.population.size)

def get_avg(country):
    return country.avg_function

def get_best(country):
    return country.best_function

vector_check_population = np.vectorize(check_population, signature='()->()')
vector_min = np.vectorize(min, signature='(),()->()')
vector_max = np.vectorize(max, signature='(),()->()')
vector_avg = np.vectorize(get_avg, signature='()->()')
vector_best = np.vectorize(get_best, signature='()->()')

class Individual:

    def __init__(self, x, function):
        self.x = x
        self.f = function(self.x)
        self.ep_n = 0

    def __lt__(self, other):
        return self.f < other.f

    def __le__(self, other):
        return self.f <= other.f

    def __gt__(self, other):
        return self.f > other.f

    def __ge__(self, other):
        return self.f >= other.f

    def __add__(self, other):
        return self.f + other.f

    def __radd__(self, other):
        return self.f + other

    def __truediv__(self, other):
        return self.f / other

    def update_function(self, function):
        self.f = function(self.x)

    @classmethod
    def crossing(cls, individual1, individual2, p, function, Xmin, Xmax):
        alpha = p
        c_min = vector_min(individual1.x, individual2.x)
        c_max = vector_max(individual1.x, individual2.x)
        I = c_max - c_min
        new_x = np.random.uniform(c_min - I * alpha, c_max + I * alpha)
        new_x = np.clip(new_x, Xmin, Xmax)
        return [cls(new_x, function)]

    def mutation(self, Xmin, Xmax, function, pmax):
        self.ep_n += 1
        self.x = self.x + pmax * np.random.uniform(-self.x, self.x) / self.ep_n
        np.clip(self.x, Xmin, Xmax)
        self.update_function(function)


class Country:

    def __init__(self, Xmin, Xmax, N, function):
        x_min = np.random.uniform(Xmin, Xmax)
        x_max = np.random.uniform(x_min, Xmax)
        self.population = []
        ind_init = np.vectorize(lambda x: Individual(x, function), signature='(n)->()')
        rand_mat = np.vectorize(lambda x: random_matrix(x, x_min, x_max), signature='()->(n)')
        v = rand_mat(np.ones(N))
        self.population = ind_init(v)
        self.sort_population()
        self.action = None
        self.enemy = None
        self.ally = None

    @property
    def best_function(self):
        return self.population[0].f

    @property
    def avg_function(self):
        return np.average(self.population)

    def update_population(self, function):
        for individual in self.population:
            individual.update_function(function)

    def sort_population(self):
        self.population = np.sort(self.population)

    def reproduction(self, n_min, n_max, p_min, p_max, f_min, f_max, ti, t_max, function, Xmin, Xmax):
        n = ceil((n_max - n_min) * (f_max - self.avg_function) / (f_max - f_min) + n_min)
        n = np.clip(n, n_min, n_max)
        p = (p_max - p_min) * (1 - ti / t_max) * (self.avg_function - f_min) / (f_max - f_min) + p_min
        p = np.clip(p, p_min, p_max)
        new_individuals = np.array([])
        for i in range(n):
            parents = np.random.choice(self.population, 2, replace=False)
            new_individuals = np.concatenate([new_individuals, Individual.crossing(parents[0], parents[1], p, function, Xmin, Xmax)])
        self.population = np.concatenate([self.population, new_individuals])
        self.sort_population()

    def extinction(self, m_min, m_max, f_min, f_max):
        m = int((m_max - m_min) * (self.avg_function - f_min) / (f_max - f_min) + m_min)
        m = m if m <= m_max else m_max
        m = m if m >= m_min else m_min
        self.population = self.population[:-m]

    def select_action(self, countries):
        self.action = random.randint(0, 3)
        if self.action == 1:
            ally_list = [country for country in countries if country.action is None and country != self]
            if ally_list:
                self.ally = ally_list.pop(random.randint(0, len(ally_list) - 1))
                self.ally.action = 1
                self.ally.ally = self
            else:
                self.action = random.choice([0, 3])
        if self.action == 2:
            enemy_list = [country for country in countries if country.action is None and country != self]
            if enemy_list:
                self.enemy = enemy_list.pop(random.randint(0, len(enemy_list) - 1))
                self.enemy.action = 2
                self.enemy.enemy = self
            else:
                self.action = random.choice([0, 3])

    def epedemic(self, elite, dead, function, Xmin, Xmax, p_max):
        n_elite = ceil(elite * len(self.population))
        n_dead = ceil(dead * len(self.population))
        self.population = self.population[:-n_dead]
        for individual in self.population[n_elite:]:
            individual.mutation(Xmin, Xmax, function, p_max)
        self.sort_population()
        self.action = None

    # def sabotage(self, n_copy):
    #     for i in range(n_copy):
    #         self.enemy.population.append(copy.copy(self.population[0]))
    #     self.action = None
    #     self.enemy = None

    def motion(self, function, Xmin, Xmax):
        x_best = self.population[0].x
        for i in range(1, len(self.population)):
            self.population[i].x = self.population[i].x + np.random.uniform(0, 2, self.population[i].x.size) * (x_best - self.population[i].x)
            np.clip(self.population[i].x, Xmin, Xmax)
        self.update_population(function)
        self.sort_population()
        self.action = None

    @staticmethod
    def trade(country1, country2, k):
        if country1.population.size <= k or country2.population.size <= k:
            k = min(country1.population.size, country2.population.size) // 2
        indexes1 = np.random.choice(country1.population.size, k, replace=False)
        indexes2 = np.random.choice(country2.population.size, k, replace=False)
        country2.population = np.concatenate([country2.population, country1.population[indexes1]])
        country1.population = np.concatenate([country1.population, country2.population[indexes2]])
        country1.population = np.delete(country1.population, indexes1)
        country2.population = np.delete(country2.population, indexes2)
        country1.sort_population()
        country2.sort_population()
        country1.action = None
        country2.action = None
        country1.ally = None
        country2.ally = None

    @staticmethod
    def war(country1, country2, l):
        if country1.population.size <= l or country2.population.size <= l:
            l = min(country1.population.size, country2.population.size)
        indexes1 = np.random.choice(country1.population.size, l, replace=False)
        indexes2 = np.random.choice(country2.population.size, l, replace=False)
        war_list1 = country1.population[indexes1]
        war_list2 = country2.population[indexes2]
        country1.population = np.delete(country1.population, indexes1)
        country2.population = np.delete(country2.population, indexes2)
        wins1 = np.where(war_list1 > war_list2)
        wins2 = np.where(war_list2 > war_list2)
        if wins1[0].size > wins2[0].size:
            np.concatenate([country1.population, war_list1])
            np.concatenate([country1.population, war_list2])
        elif wins2[0].size > wins1[0].size:
            np.concatenate([country2.population, war_list1])
            np.concatenate([country2.population, war_list2])
        else:
            np.concatenate([country1.population, war_list1])
            np.concatenate([country2.population, war_list2])
        country1.sort_population()
        country2.sort_population()
        country1.action = None
        country2.action = None
        country1.enemy = None
        country2.enemy = None


class CountriesAlgorithm_v2:

    def __init__(self, f, Xmin, Xmax, M, N, n, p, m, k, l, ep, tmax, printing=False):
        self.f = f
        self.Xmin = Xmin
        self.Xmax = Xmax
        self.n = n
        self.p = p
        self.m = m
        self.k = k
        self.l = l
        self.ep = ep
        self.tmax = tmax
        self.printing = printing
        country_init = np.vectorize(lambda x: Country(self.Xmin, self.Xmax, N, self.f))
        self.countries = country_init(np.ones(M))

    def sort(self, method):
        if method == 'avg_function':
            avg_arr = vector_avg(self.countries)
            indexes = np.argsort(avg_arr)
            self.countries = self.countries[indexes]
        else:
            avg_arr = vector_best(self.countries)
            indexes = np.argsort(avg_arr)
            self.countries = self.countries[indexes]

    def start(self):
        self._first_iteration()
        while self.ti <= self.tmax:
            stop = self._iteration()
            if stop:
                break
        return (self.result.x, self.result.f, False, self.ti)

    def testing(self, canonical_x, epsilon):
        self._first_iteration()
        while self.ti <= self.tmax:
            self._iteration()
            if np.linalg.norm(canonical_x - self.result.x) < epsilon:
                return self.result.x, self.result.f, self.ti
        return self.result.x, self.result.f, self.ti

    def _first_iteration(self):
        self.ti = 0
        self.motion = 0
        self.trade = 0
        self.war = 0
        self.epedemic = 0

    def _iteration(self):
        self.ti += 1
        for country in self.countries:
            if country.action is None:
                country.select_action(self.countries)
        for country in self.countries:
            if country.action == 0:
                self.motion += 1
                country.motion(
                    function=self.f,
                    Xmin=self.Xmin,
                    Xmax=self.Xmax
                )
            elif country.action == 1:
                self.trade += 1
                Country.trade(
                    country1=country,
                    country2=country.ally,
                    k=self.k
                )
            elif country.action == 2:
                self.war += 1
                Country.war(
                    country1=country,
                    country2=country.enemy,
                    l=self.l
                )
            elif country.action == 3:
                self.epedemic += 1
                country.epedemic(
                    elite=self.ep[0],
                    dead=self.ep[1],
                    Xmin=self.Xmin,
                    Xmax=self.Xmax,
                    function=self.f,
                    p_max=self.p[1],
                )
        if not self.countries.size:
            return True
        indexes = np.where(vector_check_population(self.countries) == True)
        self.countries = self.countries[indexes]
        try:
            self.sort('avg_function')
        except:
            return True
        f_min = self.countries[0].avg_function
        f_max = self.countries[-1].avg_function
        if f_min == f_max:
            self.sort('best_function')
            self.result = self.countries[0].population[0]
            return True
        e_individuals = []
        for country in self.countries:
            if country.population.size == 1:
                e_individuals.append(country.population[0])
                continue
            if country.population.size:
                country.reproduction(
                    n_min=self.n[0],
                    n_max=self.n[1],
                    p_min=self.p[0],
                    p_max=self.p[1],
                    f_min=f_min,
                    f_max=f_max,
                    ti=self.ti,
                    t_max=self.tmax,
                    function=self.f,
                    Xmin=self.Xmin,
                    Xmax=self.Xmax
                )
                country.extinction(
                    m_min=self.m[0],
                    m_max=self.m[1],
                    f_min=f_min,
                    f_max=f_max
                )
        indexes = np.where(vector_check_population(self.countries) == True)
        self.countries = self.countries[indexes]
        for individual in e_individuals:
            random_country = self.countries[random.randint(0, len(self.countries) - 1)]
            random_country.population = np.append(random_country.population, individual)
            random_country.sort_population()
        self.sort('best_function')
        if not self.countries.size:
            return True
        self.result = self.countries[0].population[0]

        if self.printing:
            print(
                f"{self.ti}) Лучшее решение: {self.result.x} - {self.result.f}, Стран осталось: {len(self.countries)}, Движение/Обмен/Войны/Эпидемии: {motion}/{trade}/{war}/{epedemic}")
            print(f"Общее количество особей: {sum([len(country.population) for country in self.countries])}")
        return False



import random
import numpy as np


class Plant:

    def __init__(self, x, function):
        self.x = np.array(x)
        self.f = function(self.x)
        self.function = function
    
    def calculate_n(self, f_best, f_worst, n_min, n_max):
        return round(self.f * (n_max - n_min) / (f_best - f_worst) + (f_best * n_min - f_worst * n_max) / (f_best - f_worst))

    def create_new_plant(self, sigma):
        new_x = self.x + np.random.normal(0, sigma, self.x.size)
        return Plant(new_x, self.function)

    def __lt__(self, other):
        return self.f < other.f

    def __le__(self, other):
        return self.f <= other.f

    def __gt__(self, other):
        return self.f > other.f

    def __ge__(self, other):
        return self.f >= other.f

class IWO:

    def __init__(self, s0, s_max, n_min, n_max, sigma_b, sigma_e, x_min, x_max, m, t_max, f, printing=False):
        self.s0 = s0
        self.s_max = s_max
        self.n_min = n_min
        self.n_max = n_max
        self.sigma_b = sigma_b
        self.sigma_e = sigma_e
        self.x_min = x_min
        self.x_max = x_max
        self.m = m
        self.t_max = t_max
        self.f = f
        self.printing = printing

    def _first_iteration(self):
        self.t = 0
        self.plants = []
        for i in range(self.s0):
            xi = []
            for j in range(len(self.x_min)):
                xi.append(self.x_min[j] + random.random() * (self.x_max[j] - self.x_min[j]))
            self.plants.append(Plant(xi, self.f))
        self.plants.sort()

    def _iteration(self):

        n = []
        f_best = self.plants[0].f
        f_worst = self.plants[-1].f
        for plant in self.plants:
            n.append(plant.calculate_n(f_best, f_worst, self.n_min, self.n_max))
        new_plants = []
        sigma = (((self.t_max - self.t) / self.t_max) ** self.m) * (self.sigma_b - self.sigma_e) + self.sigma_e
        for i, plant in enumerate(self.plants):
            for j in range(n[i]):
                new_plants.append(plant.create_new_plant(sigma))
        self.plants += new_plants
        self.plants.sort()
        if len(self.plants) > self.s_max:
            self.plants = self.plants[:self.s_max]
        if self.printing:
            print(self.t, self.plants[0].x, self.plants[0].f)
        self.t += 1

    def start(self):
        self._first_iteration()
        while self.t <= self.t_max:
            stop = self._iteration()
            if stop:
                break
        return (self.plants[0].x, self.plants[0].f, False, self.t)

    def testing(self, canonical_x, epsilon):
        self._first_iteration()
        while self.t <= self.t_max:
            self._iteration()
            if np.linalg.norm(canonical_x - np.array(self.plants[0].x)) < epsilon:
                return self.plants[0].x, self.plants[0].f, self.t
        return self.plants[0].x, self.plants[0].f, self.t


import numpy as np
from functools import wraps


def count_calls(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        wrapper.calls += 1
        try:
            return func(*args, **kwargs)
        except TypeError:
            return func(*args[1:], **kwargs)

    wrapper.calls = 0
    wrapper.reset = lambda: setattr(wrapper, 'calls', 0)
    return wrapper


@count_calls
def sphere(x):
    """
    Функция Сферы (Функция Де Йонга F1)

    f(x) = Σ(xi²) для i=1 до n

    Область: xi ∈ [-5.12, 5.12]
    Глобальный минимум: f(x*) = 0 в точке x* = (0, 0, ..., 0)
    """
    return np.sum(x ** 2)


@count_calls
def rastrigin(x):
    """
    Функция Растригина

    f(x) = 10n + Σ[xi² - 10cos(2πxi)] для i=1 до n

    Область: xi ∈ [-5.12, 5.12]
    Глобальный минимум: f(x*) = 0 в точке x* = (0, 0, ..., 0)
    """
    n = len(x)
    return 10 * n + np.sum(x ** 2 - 10 * np.cos(2 * np.pi * x))


@count_calls
def ackley(x):
    """
    Функция Акли

    f(x) = -20·exp(-0.2·√(1/n·Σxi²)) - exp(1/n·Σcos(2πxi)) + 20 + e

    Область: xi ∈ [-32.768, 32.768]
    Глобальный минимум: f(x*) = 0 в точке x* = (0, 0, ..., 0)
    """
    n = len(x)
    sum1 = np.sum(x ** 2)
    sum2 = np.sum(np.cos(2 * np.pi * x))
    return -20 * np.exp(-0.2 * np.sqrt(sum1 / n)) - np.exp(sum2 / n) + 20 + np.e


@count_calls
def rosenbrock(x):
    """
    Функция Розенброка (Банановая функция)

    f(x) = Σ[100(xi+1 - xi²)² + (1 - xi)²] для i=1 до n-1

    Область: xi ∈ [-5, 10]
    Глобальный минимум: f(x*) = 0 в точке x* = (1, 1, ..., 1)
    """
    return np.sum(100 * (x[1:] - x[:-1] ** 2) ** 2 + (1 - x[:-1]) ** 2)


@count_calls
def griewank(x):
    """
    Функция Гриванка

    f(x) = 1 + (1/4000)·Σxi² - Π[cos(xi/√i)] для i=1 до n

    Область: xi ∈ [-600, 600]
    Глобальный минимум: f(x*) = 0 в точке x* = (0, 0, ..., 0)
    """
    sum_sq = np.sum(x ** 2)
    prod_cos = np.prod(np.cos(x / np.sqrt(np.arange(1, len(x) + 1))))
    return sum_sq / 4000 - prod_cos + 1


@count_calls
def schwefel(x):
    """
    Функция Швефеля (Schwefel 2.26)

    f(x) = 418.9829n - Σ[xi·sin(√|xi|)] для i=1 до n

    Область: xi ∈ [-500, 500]
    Глобальный минимум: f(x*) = 0 в точке x* = (420.9687, ..., 420.9687)
    """
    n = len(x)
    return 418.9829 * n - np.sum(x * np.sin(np.sqrt(np.abs(x))))


@count_calls
def levy(x):
    """
    Функция Леви (Levy N.13)

    Область: xi ∈ [-10, 10]
    Глобальный минимум: f(x*) = 0 в точке x* = (1, 1, ..., 1)
    """
    w = 1 + (x - 1) / 4
    term1 = np.sin(np.pi * w[0]) ** 2
    term3 = (w[-1] - 1) ** 2 * (1 + np.sin(2 * np.pi * w[-1]) ** 2)
    wi = w[:-1]
    sum_term = np.sum((wi - 1) ** 2 * (1 + 10 * np.sin(np.pi * wi + 1) ** 2))
    return term1 + sum_term + term3


@count_calls
def michalewicz(x, m=10):
    """
    Функция Михалевича

    f(x) = -Σ[sin(xi)·sin^(2m)(i·xi²/π)] для i=1 до n

    Область: xi ∈ [0, π]
    Параметр m определяет крутизну (обычно m=10)
    """
    i = np.arange(1, len(x) + 1)
    return -np.sum(np.sin(x) * np.sin(i * x ** 2 / np.pi) ** (2 * m))


@count_calls
def zakharov(x):
    """
    Функция Захарова

    f(x) = Σxi² + (Σ0.5i·xi)² + (Σ0.5i·xi)⁴ для i=1 до n

    Область: xi ∈ [-5, 10]
    Глобальный минимум: f(x*) = 0 в точке x* = (0, 0, ..., 0)
    """
    sum1 = np.sum(x ** 2)
    i = np.arange(1, len(x) + 1)
    sum2 = np.sum(0.5 * i * x)
    return sum1 + sum2 ** 2 + sum2 ** 4


@count_calls
def dixonprice(x):
    """
    Функция Диксона-Прайса

    f(x) = (x1 - 1)² + Σ[i·(2xi² - xi-1)²] для i=2 до n

    Область: xi ∈ [-10, 10]
    Глобальный минимум: f(x*) = 0
    """
    i = np.arange(2, len(x) + 1)
    term1 = (x[0] - 1) ** 2
    term2 = np.sum(i * (2 * x[1:] ** 2 - x[:-1]) ** 2)
    return term1 + term2


@count_calls
def st(x):
    return np.sum((x**4 - 16 * x**2 + 5 * x) / 2)




import copy

import numpy as np
from typing import Callable, Iterable


class SandCat:
    """
    Агент песчаного кота для базового SCSO.
    """

    def __init__(self, x, function: Callable):
        self.x = np.array(x, dtype=float)
        self.function = function
        self.f = function(self.x)

    def update_cat(
            self,
            best_cat,
            r_g: float,
            ub: np.ndarray,
            lb: np.ndarray
    ):
        dim = self.x.size

        R = (2 * r_g) * (2 * np.random.rand(dim) - 1)
        abs_R = np.abs(R)

        S_r = r_g * np.random.rand(dim)

        theta = 2 * np.pi * np.random.rand(dim)

        new_x = self.x.copy()

        if np.any(abs_R > 1):
            rand_point = lb + np.random.rand(dim) * (ub - lb)
            rand_scale = np.random.rand(dim)
            new_x = S_r * (rand_point - rand_scale * self.x)
        else:
            rand_scale = np.random.rand(dim)
            direction = best_cat.x - self.x
            new_x = best_cat.x - S_r * rand_scale * direction * np.cos(theta)

        new_x = np.clip(new_x, lb, ub)
        self.x = new_x
        self.f = self.function(self.x)

    def __lt__(self, other):
        return self.f < other.f

    def __le__(self, other):
        return self.f <= other.f

    def __gt__(self, other):
        return self.f > other.f

    def __ge__(self, other):
        return self.f >= other.f


class SandCatOptimizer:

    def __init__(
            self,
            f: Callable,
            x_min: Iterable[float],
            x_max: Iterable[float],
            n_cats: int = 30,
            max_iter: int = 500,
            r_g_max: float = 2.0,
            r_g_min: float = 0.0,
            printing: bool = False
    ):
        self.objective_func = f
        self.n_cats = n_cats
        self.max_iter = max_iter

        self.lb = np.array(x_min, dtype=float)
        self.ub = np.array(x_max, dtype=float)
        self.r_g_max = r_g_max
        self.r_g_min = r_g_min

        self.printing = printing

    def _first_iteration(self):
        self.t = 0
        self.cats = []
        for _ in range(self.n_cats):
            xi = np.random.uniform(self.lb, self.ub)
            self.cats.append(SandCat(xi, self.objective_func))
        self.cats.sort()
        self.best_cat = copy.copy(self.cats[0])

    def _compute_r_g(self):
        progress = self.t / max(1, self.max_iter)
        return self.r_g_max - progress * (self.r_g_max - self.r_g_min)

    def _iteration(self):
        self.cats.sort()
        if self.cats[0].f < self.best_cat.f:
            self.best_cat = copy.copy(self.cats[0])

        r_g = self._compute_r_g()

        for cat in self.cats:
            cat.update_cat(
                best_cat=self.best_cat,
                r_g=r_g,
                ub=self.ub,
                lb=self.lb
            )

        if self.printing:
            print(self.t, self.best_cat.x, self.best_cat.f)

        self.t += 1

    def start(self):
        self._first_iteration()
        while self.t <= self.max_iter:
            self._iteration()
        return self.best_cat.x, self.best_cat.f, False, self.t

    def testing(self, canonical_x, epsilon):
        self._first_iteration()
        while self.t <= self.max_iter:
            self._iteration()
            if np.linalg.norm(canonical_x - self.best_cat.x) < epsilon:
                return self.best_cat.x, self.best_cat.f, self.t
        return self.best_cat.x, self.best_cat.f, self.t



from datetime import datetime
import random
import matplotlib.pyplot as plt


class BenchMarkParamsPlotter:

    _y_dict = {
        'calls': 'Average number of function calls',
        'avg_iterations': 'Average number of iterations',
        'success_percent': 'Convergence percentage',
        'avg_time': 'Average time',
    }

    def __init__(self, results: dict, x_param: str, method_name: str, path: str = ''):
        self.results = results
        self.x_param = x_param
        self.method_name = method_name
        self.path = path

    def plot(self):
        paths = []
        for r_key in self.results:
            x = [r.input_args.get(self.x_param) for r in self.results[r_key]]
            for key in self._y_dict:
                y = [getattr(r, key) for r in self.results[r_key]]
                plt.cla()
                plt.plot(x, y, label=f'Testing for parameter {self.x_param}')
                plt.grid()
                plt.title(f"Test function: {r_key} / Method: {self.method_name}", fontsize=14)
                plt.xlabel(self.x_param, fontsize=12)
                plt.ylabel(self._y_dict[key], fontsize=12)
                plt.legend()
                plt.tight_layout()
                img_path = f'{self.path}plot_{datetime.now().strftime("%H_%M_%S_%d_%m_%Y")}_{str(random.randint(0, 100000))}.png'
                plt.savefig(img_path)
                paths.append(img_path)
        return paths



import telebot
from telebot import apihelper
import matplotlib.pyplot as plt
import datetime
import random


class TgSender:

    _y_dict = {
        'calls': 'Среднее число вызовов функции',
        'avg_iterations': 'Среднее количество итераций',
        'success_percent': 'Процент сходимости',
        'avg_time': 'Среднее время',
    }

    def __init__(self, bot_token: str, chat_id: str, results: dict,
                 x_param: str, method_name: str, plot_names: list, path: str = ''):
        self.tag = f'#{method_name.replace(" ", "_")}_{datetime.datetime.now().strftime("%H_%M_%S_%d_%m_%Y")}'
        self.bot_token = bot_token
        self.chat_id = chat_id
        self.results = results
        self.plot_names = plot_names
        self.x_param = x_param
        self.method_name = method_name
        self.body = ''
        self.path = path
        for key, value in self.results[list(self.results.keys())[0]][0].input_args.items():
            if key != self.x_param:
                self.body += f'{key}: {value}\n'
        apihelper.READ_TIMEOUT = 120

    def create_tables(self):
        paths = []
        for r_key in self.results:
            title = f"Тестовая функция: {r_key} / Метод: {self.method_name}"
            res = []
            x = [str(r.input_args.get(self.x_param)) for r in self.results[r_key]]
            res.append(x)
            head = [self.x_param]
            for key in self._y_dict:
                head.append(key)
                y = [str(getattr(r, key)) for r in self.results[r_key]]
                res.append(y)
            res = [list(row) for row in zip(*res)]
            fig, ax = plt.subplots()
            ax.axis('off')
            table_data = [
                head,
                *res
            ]
            table = ax.table(
                cellText=table_data,
                cellLoc='center',
                loc='center'
            )
            table.auto_set_font_size(True)
            fig.tight_layout()
            img_path = f'{self.path}table_{datetime.datetime.now().strftime("%H_%M_%S_%d_%m_%Y")}_{str(random.randint(0, 100000))}.png'
            plt.savefig(img_path)
            paths.append((img_path, title))
        return paths

    def create_log_documents(self):
        result_text = ""
        for r_key in self.results:
            title = f"Тестовая функция: {r_key} / Метод: {self.method_name}"
            res = []
            x = [str(r.input_args.get(self.x_param)) for r in self.results[r_key]]
            res.append(x)
            head = [self.x_param]
            for key in self._y_dict:
                head.append(key)
                y = [str(getattr(r, key)) for r in self.results[r_key]]
                res.append(y)
            res = [list(row) for row in zip(*res)]
            table_data = [
                head,
                *res
            ]
            result_text += f"{title}\n\n"
            for row in table_data:
                for i in row:
                    result_text += f"{i:<25} "
                result_text += '\n'
        file_path = f'{self.path}log_{datetime.datetime.now().strftime("%H_%M_%S_%d_%m_%Y")}_{str(random.randint(0, 100000))}.txt'
        with open(file_path, 'w') as f:
            f.write(result_text)
        return file_path

    def send(self):
        table_paths = self.create_tables()
        log_path = self.create_log_documents()
        for table_path, title in table_paths:
            with open(table_path, 'rb') as img:
                self.send_image(img, f"{self.tag}\n{title}\n{self.body}")
        with open(log_path, 'rb') as f:
            self.send_file(f, f"#ПОЛНЫЙ_ЛОГ\n{self.body}")
        for plot_path in self.plot_names:
            with open(plot_path, 'rb') as img:
                self.send_image(img, f"{self.tag}\n{self.body}")

    def send_image(self, image, caption):
        bot = telebot.TeleBot(self.bot_token)
        bot.send_photo(
            int(self.chat_id),
            image,
            caption=caption
        )

    def send_file(self, file, caption):
        bot = telebot.TeleBot(self.bot_token)
        bot.send_document(
            int(self.chat_id),
            file,
            caption=caption
        )



import traceback


class BenchMarkParamsAggregator:

    def __init__(self, method_class: type, start_method_name: str,
                 method_kwargs: dict, param_name: str, params_values: list,
                 benchmarks: dict):
        self.param_name = param_name
        self.params_values = params_values
        self.benchmarks = benchmarks
        self.method_class = method_class
        self.start_method_name = start_method_name
        self.method_kwargs = method_kwargs

    def __call__(self, printing=True) -> dict:
        results = {}
        for key in self.benchmarks:
            results[key] = []
            for param in self.params_values:
                benchmark = self.benchmarks[key]['class'](
                    method_class=self.method_class,
                    start_method_name=self.start_method_name,
                    method_kwargs={
                        self.param_name: param,
                        **self.method_kwargs
                    },
                    **self.benchmarks[key]['params'],
                )
                try:
                    results[key].append(benchmark())
                except Exception:
                    if printing:
                        print(key)
                        traceback.print_exc()
            if printing:
                print(key, results[key])
        return results



import copy

import numpy as np
from typing import Callable, Iterable


class Bee:
    def __init__(self, x, function: Callable):
        self.x = np.array(x)
        self.function = function
        self.f = function(self.x)

    def evaluate(self):
        self.f = self.function(self.x)

    def clone_with_perturbation(self, radius, lb, ub):
        x_new = self.x + np.random.uniform(-radius, radius, size=self.x.size)
        x_new = np.clip(x_new, lb, ub)
        return Bee(x_new, self.function)

    def __lt__(self, other):
        return self.f < other.f

    def __le__(self, other):
        return self.f <= other.f

    def __gt__(self, other):
        return self.f > other.f

    def __ge__(self, other):
        return self.f >= other.f


class BeesOptimizer:

    def __init__(
            self,
            f: Callable,
            x_min: Iterable[float],
            x_max: Iterable[float],
            n_scouts: int = 30,
            n_selected: int = 10,
            n_elite: int = 3,
            n_recruited_elite: int = 10,
            n_recruited_selected: int = 5,
            max_iter: int = 500,
            radius: float = 0.1,
            radius_shrink: float = 0.95,
            printing: bool = False
    ):
        self.objective_func = f
        self.lb = np.array(x_min, dtype=float)
        self.ub = np.array(x_max, dtype=float)

        self.n_scouts = n_scouts
        self.n_selected = min(n_selected, n_scouts)
        self.n_elite = min(n_elite, self.n_selected)
        self.n_recruited_elite = n_recruited_elite
        self.n_recruited_selected = n_recruited_selected

        self.max_iter = max_iter
        self.radius = radius
        self.radius_shrink = radius_shrink
        self.printing = printing

    def _random_bee(self):
        x = np.random.uniform(self.lb, self.ub)
        return Bee(x, self.objective_func)

    def _first_iteration(self):
        self.t = 0
        self.bees = [self._random_bee() for _ in range(self.n_scouts)]
        self.bees.sort()
        self.best = copy.copy(self.bees[0])

    def _iteration(self):
        self.bees.sort()
        if self.bees[0].f < self.best.f:
            self.best = copy.copy(self.bees[0])

        selected_sites = self.bees[:self.n_selected]
        elite_sites = selected_sites[:self.n_elite]
        other_selected = selected_sites[self.n_elite:]

        new_bees = []

        for site in elite_sites:
            local_bees = [site]
            for _ in range(self.n_recruited_elite - 1):
                local_bees.append(site.clone_with_perturbation(
                    self.radius, self.lb, self.ub
                ))
            local_bees.sort()
            new_bees.append(local_bees[0])

        for site in other_selected:
            local_bees = [site]
            for _ in range(self.n_recruited_selected - 1):
                local_bees.append(site.clone_with_perturbation(
                    self.radius, self.lb, self.ub
                ))
            local_bees.sort()
            new_bees.append(local_bees[0])

        n_new = self.n_scouts - len(new_bees)
        for _ in range(n_new):
            new_bees.append(self._random_bee())

        self.bees = new_bees
        if self.printing:
            print(self.t, self.best.x, self.best.f)

        self.radius *= self.radius_shrink

        self.t += 1

    def start(self):
        self._first_iteration()
        while self.t <= self.max_iter:
            self._iteration()
        return self.best.x, self.best.f, False, self.t

    def testing(self, canonical_x, epsilon):
        self._first_iteration()
        while self.t <= self.max_iter:
            self._iteration()
            if np.linalg.norm(canonical_x - self.best.x) < epsilon:
                return self.best.x, self.best.f, self.t
        return self.best.x, self.best.f, self.t



import random
import numpy as np
from operator import attrgetter
from math import ceil, cos, sin


class Individual:

    def __init__(self, x, function):
        self.x = x
        self.f = function(self.x)
        self.ep_n = 0

    def update_function(self, function):
        self.f = function(self.x)

    @classmethod
    def crossing(cls, individual1, individual2, p, function, Xmin, Xmax):
        x1 = []
        x2 = []
        for i in range(len(individual1.x)):
            if i % 2:
                x1.append(individual1.x[i])
                x2.append(individual2.x[i])
            else:
                x1.append(individual2.x[i])
                x2.append(individual1.x[i])
            x1[i] += random.uniform(-p * x1[i], p * x1[i])
            x1[i] = x1[i] if x1[i] >= Xmin[i] else Xmin[i]
            x1[i] = x1[i] if x1[i] <= Xmax[i] else Xmax[i]
            x2[i] += random.uniform(-p * x2[i], p * x2[i])
            x2[i] = x2[i] if x2[i] >= Xmin[i] else Xmin[i]
            x2[i] = x2[i] if x2[i] <= Xmax[i] else Xmax[i]
        return [cls(x1, function), cls(x2, function)]

    @classmethod
    def crossing2(cls, individual1, individual2, p, function, Xmin, Xmax):
        x1 = []
        x2 = []
        random_list = list(range(len(individual1.x)))
        random.shuffle(random_list)
        random_list = random_list[:len(individual1.x) // 2]
        for i in range(len(individual1.x)):
            if i in random_list:
                x1.append(individual1.x[i])
                x2.append(individual2.x[i])
            else:
                x1.append(individual2.x[i])
                x2.append(individual1.x[i])
            x1[i] += random.uniform(-p * x1[i], p * x1[i])
            x1[i] = x1[i] if x1[i] >= Xmin[i] else Xmin[i]
            x1[i] = x1[i] if x1[i] <= Xmax[i] else Xmax[i]
            x2[i] += random.uniform(-p * x2[i], p * x2[i])
            x2[i] = x2[i] if x2[i] >= Xmin[i] else Xmin[i]
            x2[i] = x2[i] if x2[i] <= Xmax[i] else Xmax[i]
        return [cls(x1, function), cls(x2, function)]

    @classmethod
    def crossing3(cls, individual1, individual2, p, function, Xmin, Xmax):
        x1 = []
        x2 = []
        max_n = len(individual1.x) // 2
        n = random.randint(1, max_n)
        l = []
        for i in range(len(individual1.x) // n):
            l += [j + 2 * i * n for j in range(n) if j + 2 * i * n < len(individual1.x)]
        for i in range(len(individual1.x)):
            if i in l:
                x1.append(individual1.x[i])
                x2.append(individual2.x[i])
            else:
                x1.append(individual2.x[i])
                x2.append(individual1.x[i])
            x1[i] += random.uniform(-p * x1[i], p * x1[i])
            x1[i] = x1[i] if x1[i] >= Xmin[i] else Xmin[i]
            x1[i] = x1[i] if x1[i] <= Xmax[i] else Xmax[i]
            x2[i] += random.uniform(-p * x2[i], p * x2[i])
            x2[i] = x2[i] if x2[i] >= Xmin[i] else Xmin[i]
            x2[i] = x2[i] if x2[i] <= Xmax[i] else Xmax[i]
        return [cls(x1, function), cls(x2, function)]

    @classmethod
    def crossing4(cls, individual1, individual2, p, function, Xmin, Xmax):
        x1 = []
        x2 = []
        for i in range(len(individual1.x)):
            x1.append((individual1.x[i] + individual2.x[i]) / 2)
            x2.append((individual1.x[i] + individual2.x[i]) / 2)
            x2[i] += random.uniform(-p * x2[i], p * x2[i])
            x2[i] = x2[i] if x2[i] >= Xmin[i] else Xmin[i]
            x2[i] = x2[i] if x2[i] <= Xmax[i] else Xmax[i]
        return [cls(x1, function), cls(x2, function)]

    @classmethod
    def crossing5(cls, individual1, individual2, p, function, Xmin, Xmax):
        x1 = []
        # x2 = []
        for i in range(len(individual1.x)):
            x1.append((individual1.x[i] + individual2.x[i]) / 2)
            # x2.append((individual1.x[i] + individual2.x[i]) / 2)
            # x2[i] += random.uniform(-p * x2[i], p * x2[i])
            # x2[i] = x2[i] if x2[i] >= Xmin[i] else Xmin[i]
            # x2[i] = x2[i] if x2[i] <= Xmax[i] else Xmax[i]
        return [cls(x1, function)] #, cls(x2, function)]

    @classmethod
    def crossing6(cls, individual1, individual2, p, function, Xmin, Xmax):
        x1 = []
        k = random.random()
        if k > 0.5:
            x2 = []
        for i in range(len(individual1.x)):
            x1.append((individual1.x[i] + individual2.x[i]) / 2)
            if k > 0.5:
                x2.append((individual1.x[i] + individual2.x[i]) / 2)
                x2[i] += random.uniform(-p * x2[i], p * x2[i])
                x2[i] = x2[i] if x2[i] >= Xmin[i] else Xmin[i]
                x2[i] = x2[i] if x2[i] <= Xmax[i] else Xmax[i]
        if k > 0.5:
            return [cls(x1, function), cls(x2, function)]
        return [cls(x1, function)]

    @classmethod
    def crossing7(cls, individual1, individual2, p, function, Xmin, Xmax):
        k = random.random()
        if k <= 1 / 3:
            return Individual.crossing4(individual1, individual2, p, function, Xmin, Xmax)
        elif k <= 2 / 3:
            return Individual.crossing8(individual1, individual2, p, function, Xmin, Xmax)
        else:
            return Individual.crossing6(individual1, individual2, p, function, Xmin, Xmax)

    @classmethod
    def crossing8(cls, individual1, individual2, p, function, Xmin, Xmax):
        new_x = []
        alpha = 0.5
        for i in range(len(individual1.x)):
            c_min = min(individual1.x[i], individual2.x[i])
            c_max = max(individual1.x[i], individual2.x[i])
            I = c_max - c_min
            new_x.append(random.uniform(c_min - I * alpha, c_max + I * alpha))
        new_x = np.clip(new_x, Xmin, Xmax)
        return [cls(new_x, function)]


    @classmethod
    def crossing9(cls, individual1, individual2, p, function, Xmin, Xmax):
        new_x = []
        alpha = 0.5
        for i in range(len(individual1.x)):
            c_min = min(individual1.x[i], individual2.x[i])
            c_max = max(individual1.x[i], individual2.x[i])
            I = c_max - c_min
            new_x.append(random.uniform(c_min - I * alpha, c_max + I * alpha))
        new_x = np.clip(new_x, Xmin, Xmax)
        if random.random() <= 0.5:
            for i in range(len(individual1.x)):
                new_x[i] += random.uniform(-p * new_x[i], p * new_x[i])
        new_x = np.clip(new_x, Xmin, Xmax)
        return [cls(new_x, function)]

    @classmethod
    def crossing10(cls, individual1, individual2, p, function, Xmin, Xmax, p2):
        k = random.random()
        if k <= 1 / 3:
            return Individual.crossing9(individual1, individual2, p, function, Xmin, Xmax, p2)
        else:
            return Individual.crossing8(individual1, individual2, p, function, Xmin, Xmax)

    @classmethod
    def crossing11(cls, individual1, individual2, p, function, Xmin, Xmax):
        new_x = []
        alpha = p
        for i in range(len(individual1.x)):
            c_min = min(individual1.x[i], individual2.x[i])
            c_max = max(individual1.x[i], individual2.x[i])
            I = c_max - c_min
            new_x.append(random.uniform(c_min - I * alpha, c_max + I * alpha))
        new_x = np.clip(new_x, Xmin, Xmax)
        return [cls(new_x, function)]

    def mutation(self, Xmin, Xmax, function, pmax):
        self.ep_n += 1
        for i in range(len(self.x)):
            self.x[i] += pmax * random.uniform(-self.x[i], self.x[i]) / self.ep_n
            self.x[i] = self.x[i] if self.x[i] >= Xmin[i] else Xmin[i]
            self.x[i] = self.x[i] if self.x[i] <= Xmax[i] else Xmax[i]
        self.update_function(function)


class Country:

    def __init__(self, Xmin, Xmax, N, function):
        x_min = []
        x_max = []
        for i in range(len(Xmin)):
            x_min.append(random.uniform(Xmin[i], Xmax[i]))
            x_max.append(random.uniform(x_min[i], Xmax[i]))
        self.population = []
        for i in range(N):
            x = []
            for j in range(len(Xmin)):
                x.append(random.uniform(x_min[j], x_max[j]))
            self.population.append(Individual(np.array(x), function))
        self.sort_population()
        self.action = None
        self.enemy = None
        self.ally = None

    @property
    def best_function(self):
        return self.population[0].f

    @property
    def avg_function(self):
        return sum([individual.f for individual in self.population]) / len(self.population)

    def update_population(self, function):
        for individual in self.population:
            individual.update_function(function)

    def sort_population(self):
        self.population.sort(key=attrgetter('f'))

    def reproduction(self, n_min, n_max, p_min, p_max, f_min, f_max, ti, t_max, function, Xmin, Xmax):
        n = ceil((n_max - n_min) * (f_max - self.avg_function) / (f_max - f_min) + n_min)
        n = np.clip(n, n_min, n_max)
        p = (p_max - p_min) * (1 - ti / t_max) * (self.avg_function - f_min) / (f_max - f_min) + p_min
        p = np.clip(p, p_min, p_max)
        # p2 = (1 - ti / t_max) * (self.avg_function - f_min) / (f_max - f_min)
        new_individuals = []

        for i in range(n):
            if len(self.population) == 2 and self.population[0] == self.population[1]:
                new_individuals.extend(Individual.crossing11(self.population[0], self.population[1], p, function, Xmin, Xmax))
                continue
            k1 = random.randint(0, len(self.population) - 1)
            individual1 = self.population[k1]
            k2 = k1
            while k2 == k1:
                k2 = random.randint(0, len(self.population) - 1)
            individual2 = self.population[k2]
            new_individuals.extend(Individual.crossing11(individual1, individual2, p, function, Xmin, Xmax))
        self.population.extend(new_individuals)
        self.sort_population()

    def extinction(self, m_min, m_max, f_min, f_max):
        m = int((m_max - m_min) * (self.avg_function - f_min) / (f_max - f_min) + m_min)
        m = m if m <= m_max else m_max
        m = m if m >= m_min else m_min
        self.population = self.population[:-m]

    def select_action(self, countries):
        self.action = random.randint(0, 3)
        if self.action == 1:
            ally_list = [country for country in countries if country.action is None and country != self]
            if ally_list:
                self.ally = ally_list.pop(random.randint(0, len(ally_list) - 1))
                self.ally.action = 1
                self.ally.ally = self
            else:
                self.action = random.choice([0, 3])
        if self.action == 2:
            enemy_list = [country for country in countries if country.action is None and country != self]
            if enemy_list:
                self.enemy = enemy_list.pop(random.randint(0, len(enemy_list) - 1))
                self.enemy.action = 2
                self.enemy.enemy = self
            else:
                self.action = random.choice([0, 3])

    def epedemic(self, elite, dead, function, Xmin, Xmax, p_max):
        n_elite = ceil(elite * len(self.population))
        n_dead = ceil(dead * len(self.population))
        self.population = self.population[:-n_dead]
        for individual in self.population[n_elite:]:
            individual.mutation(Xmin, Xmax, function, p_max)
        self.sort_population()
        self.action = None

    # def sabotage(self, n_copy):
    #     for i in range(n_copy):
    #         self.enemy.population.append(copy.copy(self.population[0]))
    #     self.action = None
    #     self.enemy = None

    def motion(self, function, Xmin, Xmax):
        x_best = self.population[0].x
        for i in range(1, len(self.population)):
            for j in range(len(x_best)):
                self.population[i].x[j] += random.uniform(0, 2) * (x_best[j] - self.population[i].x[j])
                self.population[i].x[j] = self.population[i].x[j] if self.population[i].x[j] >= Xmin[j] else Xmin[j]
                self.population[i].x[j] = self.population[i].x[j] if self.population[i].x[j] <= Xmax[j] else Xmax[j]
        self.update_population(function)
        self.sort_population()
        self.action = None

    @staticmethod
    def trade(country1, country2, k):
        trade_list1 = []
        trade_list2 = []
        if len(country1.population) <= k or len(country2.population) <= k:
            k = min(len(country1.population), len(country2.population)) // 2
        for i in range(k):
            trade_list1.append(country1.population.pop(random.randint(0, len(country1.population) - 1)))
            trade_list2.append(country2.population.pop(random.randint(0, len(country2.population) - 1)))
        country1.population.extend(trade_list2)
        country2.population.extend(trade_list1)
        country1.sort_population()
        country2.sort_population()
        country1.action = None
        country2.action = None
        country1.ally = None
        country2.ally = None

    @staticmethod
    def war(country1, country2, l):
        war_list1 = []
        war_list2 = []
        if len(country1.population) <= l or len(country2.population) <= l:
            l = min(len(country1.population), len(country2.population))
        for i in range(l):
            war_list1.append(country1.population.pop(random.randint(0, len(country1.population) - 1)))
            war_list2.append(country2.population.pop(random.randint(0, len(country2.population) - 1)))
        wins1 = 0
        wins2 = 0
        for i in range(l-1, -1, -1):
            if war_list1[i].f < war_list2[i].f:
                war_list2.pop(i)
                wins1 += 1
            elif war_list1[i].f > war_list2[i].f:
                war_list1.pop(i)
                wins2 += 1
        if wins1 > wins2:
            country1.population.extend(war_list1)
            country1.population.extend(war_list2)
        elif wins2 > wins1:
            country2.population.extend(war_list1)
            country2.population.extend(war_list2)
        else:
            country1.population.extend(war_list1)
            country2.population.extend(war_list2)
        country1.sort_population()
        country2.sort_population()
        country1.action = None
        country2.action = None
        country1.enemy = None
        country2.enemy = None


class CountriesAlgorithm:

    def __init__(self, f, Xmin, Xmax, M, N, n, p, m, k, l, ep, tmax, printing=False, memory_list=None):
        self.f = f
        self.Xmin = Xmin
        self.Xmax = Xmax
        self.n = n
        self.p = p
        self.m = m
        self.k = k
        self.l = l
        self.ep = ep
        self.tmax = tmax
        self.countries = []
        self.printing = printing
        self.memory_list = memory_list
        for i in range(M):
            self.countries.append(Country(self.Xmin, self.Xmax, N, self.f))

    def start(self):
        self._first_iteration()
        while self.ti <= self.tmax:
            stop = self._iteration()
            if stop:
                break
        return (self.result.x, self.result.f, False, self.ti)

    def testing(self, canonical_x, epsilon):
        self._first_iteration()
        t = 0
        while self.ti <= self.tmax:
            self._iteration()
            t += 1
            if np.linalg.norm(canonical_x - self.result.x) < epsilon:
                return self.result.x, self.result.f, self.ti
        return self.result.x, self.result.f, self.ti

    def _first_iteration(self):
        self.ti = 0
        self.motion = 0
        self.trade = 0
        self.war = 0
        self.epedemic = 0
        if self.memory_list is not None:
            self.memory_list[0] = False

    def _iteration(self):
        self.ti += 1
        for country in self.countries:
            if country.action is None:
                country.select_action(self.countries)
        for country in self.countries:
            if country.action == 0:
                self.motion += 1
                country.motion(
                    function=self.f,
                    Xmin=self.Xmin,
                    Xmax=self.Xmax
                )
            elif country.action == 1:
                self.trade += 1
                Country.trade(
                    country1=country,
                    country2=country.ally,
                    k=self.k
                )
            elif country.action == 2:
                self.war += 1
                Country.war(
                    country1=country,
                    country2=country.enemy,
                    l=self.l
                )
            elif country.action == 3:
                self.epedemic += 1
                country.epedemic(
                    elite=self.ep[0],
                    dead=self.ep[1],
                    Xmin=self.Xmin,
                    Xmax=self.Xmax,
                    function=self.f,
                    p_max=self.p[1],
                )
        self.countries = [country for country in self.countries if country.population]
        self.countries = sorted(self.countries, key=attrgetter('avg_function'))
        if not self.countries:
            return True
        f_min = self.countries[0].avg_function
        f_max = self.countries[-1].avg_function
        if f_min == f_max:
            self.countries = sorted(self.countries, key=attrgetter('best_function'))
            self.result = self.countries[0].population[0]
            return True
        e_individuals = []
        for country in self.countries:
            if len(country.population) == 1:
                e_individuals.append(country.population[0])
                continue
            if country.population:
                country.reproduction(
                    n_min=self.n[0],
                    n_max=self.n[1],
                    p_min=self.p[0],
                    p_max=self.p[1],
                    f_min=f_min,
                    f_max=f_max,
                    ti=self.ti,
                    t_max=self.tmax,
                    function=self.f,
                    Xmin=self.Xmin,
                    Xmax=self.Xmax
                )
                country.extinction(
                    m_min=self.m[0],
                    m_max=self.m[1],
                    f_min=f_min,
                    f_max=f_max
                )
        self.countries = [country for country in self.countries if country.population]
        for individual in e_individuals:
            random_country = self.countries[random.randint(0, len(self.countries) - 1)]
            random_country.population.append(individual)
            random_country.sort_population()
        self.countries = sorted(self.countries, key=attrgetter('best_function'))
        if not self.countries:
            return True
        self.result = self.countries[0].population[0]

        if self.printing:
            print(
                f"{self.ti}) Лучшее решение: {self.result.x} - {self.result.f}, Стран осталось: {len(self.countries)}, Движение/Обмен/Войны/Эпидемии: {self.motion}/{self.trade}/{self.war}/{self.epedemic}")
            print(f"Общее количество особей: {sum([len(country.population) for country in self.countries])}")

        if self.memory_list is not None:
            self.memory_list[0] = self.ti
            for i in range(len(self.result.x)):
                self.memory_list[i + 1] = float(self.result.x[i])
            self.memory_list[-1] = float(self.result.f)
        return False
